/*
 * qlogger_console.cpp
 *
 *  Created on: May 18, 2013
 *      Author: root
 */

#include "../Includes/qstdinc.h"
#include "../Includes/qsync.h"
#include "../Includes/qlogger.h"
#include "../Includes/qtime.h"
#include "../Includes/qsock.h"

QLOG_REGISTER_SOURCE

#ifndef _WIN32
static int QLogConsoleSock(uint16_t nPort)
#else
static SOCKET QLogConsoleSock(uint16_t nPort)
#endif
{
	/* Create socket */
#ifndef _WIN32
	int nSock = socket ( AF_INET, SOCK_STREAM, IPPROTO_TCP );
	if (nSock < 0)
#else
	SOCKET nSock = socket ( AF_INET, SOCK_STREAM, IPPROTO_TCP );
	if (nSock == INVALID_SOCKET)
#endif
	{
		QLogInfo(0, "Log console socket creation failed %d", errno);
		return nSock;
	}

	/* Setup socket */
	// Reuse
	long bReuse = 1 ;
	int nRetVal = setsockopt ( nSock, SOL_SOCKET, SO_REUSEADDR, ( char* ) &bReuse, sizeof ( bReuse ) );
	if ( nRetVal != 0 )
	{
		QLogInfo(0, "Log console socket / SO_REUSEADDR %d", errno);
#ifndef _WIN32
		close(nSock);
		return -1;
#else
		closesocket(nSock);
		return INVALID_SOCKET;
#endif
	}

	 // Bind
	sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr  = inet_addr("127.0.0.1");
	addr.sin_port = htons ( nPort );

	nRetVal = bind ( nSock, ( sockaddr* ) &addr, sizeof ( addr ) );
	if ( nRetVal != 0 )
	{
		QLogInfo(0, "Log console socket / bind %d", errno);
#ifndef _WIN32
		close(nSock);
		return -1;
#else
		closesocket(nSock);
		return INVALID_SOCKET;
#endif
	}

	return nSock;
}

#ifndef _WIN32
static void* QLogConsoleHandler(void* pData)
#else
static DWORD WINAPI QLogConsoleHandler(LPVOID pData)
#endif
{
	/* Convert to socket */
	int nSock = (int)(int64_t)pData;
	QTCPSock iosock(nSock);

	/* Handle commands */
	while (iosock.IsOpen())
	{
	}

	return NULL;
}

static void QLogStartNewConsole(int64_t nSock)
{
#ifndef _WIN32
	pthread_t thread;
	pthread_attr_t pattr;
	pthread_attr_init ( &pattr );
	pthread_attr_setdetachstate ( &pattr, PTHREAD_CREATE_DETACHED );
	pthread_create ( &thread, &pattr, QLogConsoleHandler, (void*)nSock );
	pthread_attr_destroy ( &pattr );
#else
	CreateThread(NULL, 0, QLogConsoleHandler, NULL, 0, (void*)nSock);
#endif
}

#ifndef _WIN32
static void* QLogConsoleListener(void* pData)
#else
static DWORD WINAPI QLogConsoleListener(LPVOID pData)
#endif
{
	/* Convert to uint16 */
	uint16_t nPort = (uint16_t)(uint64_t)pData;

	/* Listen in loop */
	while (true)
	{
		/* Try to create socket */
#ifndef _WIN32
		int nLSock = QLogConsoleSock(nPort);
		if (nLSock < 0)
#else
		SOCKET nLSock = QLogConsoleSock(nPort);
		if (nLSock == INVALID_SOCKET)
#endif
		{
			QLogInfo(0, "Waiting before trying to create socket..");
#ifndef _WIN32
			sleep(1);
#else
			::Sleep(1000);
#endif
			continue;
		}

		/* Listen and accept incoming connections */
		sockaddr_in addr2;
#ifndef _WIN32
		socklen_t addrlen = sizeof ( addr2 );
#else
		int addrlen = sizeof ( addr2 );
#endif

		/* Do listen */
#ifndef _WIN32
		int nRetVal = listen (nLSock, SOMAXCONN );
		if (nRetVal < 0)
#else
		SOCKET nRetVal = listen (nLSock, SOMAXCONN );
		if (nRetVal == INVALID_SOCKET)
#endif
		{
			QLogError(0, "Log console listen failed with code: %d", errno);
#ifndef _WIN32
			close(nLSock);
#else
			closesocket(nLSock);
#endif
			nLSock = -1;
			continue;
		}

		/* Do accept */
		while(true)
		{
			/* Try to accept connection */
			int64_t nRSock = accept ( nLSock, ( sockaddr* ) &addr2, &addrlen );
			if (nRSock < 0)
			{
				QLogError(0, "Log console accept failed with code: %d", errno);
#ifndef _WIN32
				close(nLSock);
#else
				closesocket(nLSock);
#endif
				nLSock = -1;
				break;
			}

			/* Start new thread */
			QLogInfo(0, "Log console accepted new connection...");
			QLogStartNewConsole(nRSock);
		}
	}

	return NULL;
}

bool QLogConsoleStart(uint16_t nPort)
{
#ifndef _WIN32
	/* Create listener thread */
    pthread_t thread;
    pthread_attr_t pattr;
    pthread_attr_init ( &pattr );
    pthread_attr_setdetachstate ( &pattr, PTHREAD_CREATE_DETACHED );
    pthread_create ( &thread, &pattr, QLogConsoleListener, NULL );
    pthread_attr_destroy ( &pattr );
#else
	CreateThread(NULL, 0, QLogConsoleListener, NULL, 0, NULL);
#endif

	/* Listener thread created, relax */
	QLogInfo(0, "Started log console listener on port %d", (int)nPort);
	return true;
}
